# Distributed under the terms of the Modified BSD License.
import copy
import sys
import base64
from hdijupyterutils.constants import (
    EVENTS_HANDLER_CLASS_NAME,
    LOGGING_CONFIG_CLASS_NAME,
)
from hdijupyterutils.utils import join_paths
from hdijupyterutils.configuration import override as _override
from hdijupyterutils.configuration import override_all as _override_all
from hdijupyterutils.configuration import with_override

from .constants import (
    HOME_PATH,
    CONFIG_FILE,
    MAGICS_LOGGER_NAME,
    LIVY_KIND_PARAM,
    LANG_SCALA,
    LANG_PYTHON,
    LANG_R,
    SESSION_KIND_SPARKR,
    SESSION_KIND_SPARK,
    SESSION_KIND_PYSPARK,
    CONFIGURABLE_RETRY,
    NO_AUTH,
    AUTH_BASIC,
)
from sparkmagic.livyclientlib.exceptions import BadUserConfigurationException

# import sparkmagic.utils.constants as constants


from requests_kerberos import REQUIRED


d = {}
path = join_paths(HOME_PATH, CONFIG_FILE)


def override(config, value):
    _override(d, path, config, value)


def override_all(obj):
    _override_all(d, obj)


_with_override = with_override(d, path)


# Helpers


def get_livy_kind(language):
    if language == LANG_SCALA:
        return SESSION_KIND_SPARK
    elif language == LANG_PYTHON:
        return SESSION_KIND_PYSPARK
    elif language == LANG_R:
        return SESSION_KIND_SPARKR
    else:
        raise BadUserConfigurationException(
            "Cannot get session kind for {}.".format(language)
        )


def get_auth_value(username, password):
    if username == "" and password == "":
        return NO_AUTH
    return AUTH_BASIC


@_with_override
def authenticators():
    return {
        "Kerberos": "sparkmagic.auth.kerberos.Kerberos",
        "None": "sparkmagic.auth.customauth.Authenticator",
        "Basic_Access": "sparkmagic.auth.basic.Basic",
    }


# Configs


def get_session_properties(language):
    properties = copy.deepcopy(session_configs())
    properties[LIVY_KIND_PARAM] = get_livy_kind(language)
    return properties


@_with_override
def session_configs():
    return {}


@_with_override
def kernel_python_credentials():
    return {
        "username": "",
        "base64_password": "",
        "url": "http://localhost:8998",
        "auth": NO_AUTH,
    }


def base64_kernel_python_credentials():
    return _credentials_override(kernel_python_credentials)


# No one's gonna use pyspark and pyspark3 notebook on different endpoints. Reuse the old config.
@_with_override
def kernel_python3_credentials():
    return kernel_python_credentials()


def base64_kernel_python3_credentials():
    return base64_kernel_python_credentials()


@_with_override
def kernel_scala_credentials():
    return {
        "username": "",
        "base64_password": "",
        "url": "http://localhost:8998",
        "auth": NO_AUTH,
    }


def base64_kernel_scala_credentials():
    return _credentials_override(kernel_scala_credentials)


@_with_override
def kernel_r_credentials():
    return {
        "username": "",
        "base64_password": "",
        "url": "http://localhost:8998",
        "auth": NO_AUTH,
    }


def base64_kernel_r_credentials():
    return _credentials_override(kernel_r_credentials)


@_with_override
def logging_config():
    return {
        "version": 1,
        "formatters": {
            "magicsFormatter": {
                "format": "%(asctime)s\t%(levelname)s\t%(message)s",
                "datefmt": "",
            }
        },
        "handlers": {
            "magicsHandler": {
                "class": LOGGING_CONFIG_CLASS_NAME,
                "formatter": "magicsFormatter",
                "home_path": HOME_PATH,
            }
        },
        "loggers": {
            MAGICS_LOGGER_NAME: {
                "handlers": ["magicsHandler"],
                "level": "DEBUG",
                "propagate": 0,
            }
        },
    }


@_with_override
def events_handler_class():
    return EVENTS_HANDLER_CLASS_NAME


@_with_override
def wait_for_idle_timeout_seconds():
    return 15


@_with_override
def livy_session_startup_timeout_seconds():
    return 60


@_with_override
def fatal_error_suggestion():
    return """The code failed because of a fatal error:
\t{}.

Some things to try:
a) Make sure Spark has enough available resources for Jupyter to create a Spark context.
b) Contact your Jupyter administrator to make sure the Spark magics library is configured correctly.
c) Restart the kernel."""


@_with_override
def resource_limit_mitigation_suggestion():
    return ""


@_with_override
def ignore_ssl_errors():
    return False


@_with_override
def coerce_dataframe():
    return True


@_with_override
def use_auto_viz():
    return True


@_with_override
def default_maxrows():
    return 2500


@_with_override
def default_samplemethod():
    return "take"


@_with_override
def default_samplefraction():
    return 0.1


@_with_override
def pyspark_dataframe_encoding():
    return "utf-8"


@_with_override
def heartbeat_refresh_seconds():
    return 30


@_with_override
def heartbeat_retry_seconds():
    return 10


@_with_override
def livy_server_heartbeat_timeout_seconds():
    return 60


@_with_override
def server_extension_default_kernel_name():
    return "pysparkkernel"


@_with_override
def custom_headers():
    return {}


@_with_override
def retry_policy():
    return CONFIGURABLE_RETRY


@_with_override
def retry_seconds_to_sleep_list():
    return [0.2, 0.5, 1, 3, 5]


@_with_override
def configurable_retry_policy_max_retries():
    # Sum of default values is ~10 seconds.
    # Plus 15 seconds more wanted, that's 3 more 5 second retries.
    return 8


@_with_override
def shutdown_session_on_spark_statement_errors():
    # If set to true, any spark statement errors will cause the Livy
    # session to be cleaned up
    return False


@_with_override
def all_errors_are_fatal():
    # If set to true, any error will be considered fatal and be raised
    return False


@_with_override
def cleanup_all_sessions_on_exit():
    # If set to true, all registered livy sessions will be attempted to be cleaned up before exiting
    return False


@_with_override
def kerberos_auth_configuration():
    return {"mutual_authentication": REQUIRED}


def _credentials_override(f):
    """Provides special handling for credentials. It still calls _override().
    If 'base64_password' in config is set, it will base64 decode it and returned in return value's 'password' field.
    If 'base64_password' is not set, it will fallback to 'password' in config.
    """
    credentials = f()
    base64_decoded_credentials = {
        k: credentials.get(k) for k in ("username", "password", "url", "auth")
    }
    base64_password = credentials.get("base64_password")
    if base64_password is not None:
        try:
            base64_decoded_credentials["password"] = base64.b64decode(
                base64_password
            ).decode()
        except Exception:
            exception_type, exception, traceback = sys.exc_info()
            msg = "base64_password for %s contains invalid base64 string: %s %s" % (
                f.__name__,
                exception_type,
                exception,
            )
            raise BadUserConfigurationException(msg)
    if base64_decoded_credentials["auth"] is None:
        base64_decoded_credentials["auth"] = get_auth_value(
            base64_decoded_credentials["username"],
            base64_decoded_credentials["password"],
        )
    return base64_decoded_credentials



# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.
import json
from time import sleep
import requests
import sparkmagic.utils.configuration as conf
from sparkmagic.utils.sparklogger import SparkLog
from .exceptions import HttpClientException


class ReliableHttpClient(object):
    """Http client that is reliable in its requests. Uses requests library."""

    def __init__(self, endpoint, headers, retry_policy):
        self._endpoint = endpoint
        self._headers = headers
        self._retry_policy = retry_policy
        self._auth = self._endpoint.auth
        self._session = requests.Session()
        self.logger = SparkLog("ReliableHttpClient")
        self.verify_ssl = not conf.ignore_ssl_errors()
        if not self.verify_ssl:
            self.logger.debug(
                "ATTENTION: Will ignore SSL errors. This might render you vulnerable to attacks."
            )
            requests.packages.urllib3.disable_warnings()

    def get_headers(self):
        return self._headers

    def compose_url(self, relative_url):
        r_u = "/{}".format(relative_url.rstrip("/").lstrip("/"))
        return self._endpoint.url + r_u

    def get(self, relative_url, accepted_status_codes):
        """Sends a get request. Returns a response."""
        return self._send_request(
            relative_url, accepted_status_codes, self._session.get
        )

    def post(self, relative_url, accepted_status_codes, data):
        """Sends a post request. Returns a response."""
        return self._send_request(
            relative_url, accepted_status_codes, self._session.post, data
        )

    def delete(self, relative_url, accepted_status_codes):
        """Sends a delete request. Returns a response."""
        return self._send_request(
            relative_url, accepted_status_codes, self._session.delete
        )

    def _send_request(self, relative_url, accepted_status_codes, function, data=None):
        return self._send_request_helper(
            self.compose_url(relative_url), accepted_status_codes, function, data, 0
        )

    def _send_request_helper(
        self, url, accepted_status_codes, function, data, retry_count
    ):
        while True:
            try:
                if data is None:
                    r = function(
                        url,
                        headers=self._headers,
                        auth=self._auth,
                        verify=self.verify_ssl,
                    )
                else:
                    r = function(
                        url,
                        headers=self._headers,
                        auth=self._auth,
                        data=json.dumps(data),
                        verify=self.verify_ssl,
                    )
            except requests.exceptions.RequestException as e:
                error = True
                r = None
                status = None
                text = None

                self.logger.error("Request to '{}' failed with '{}'".format(url, e))
            else:
                error = False
                status = r.status_code
                text = r.text

            if error or status not in accepted_status_codes:
                if self._retry_policy.should_retry(status, error, retry_count):
                    sleep(self._retry_policy.seconds_to_sleep(retry_count))
                    retry_count += 1
                    continue

                if error:
                    raise HttpClientException(
                        "Error sending http request and maximum retry encountered."
                    )
                else:
                    raise HttpClientException(
                        "Invalid status code '{}' from {} with error payload: {}".format(
                            status, url, text
                        )
                    )
            return r



# -*- coding: utf-8 -*-
import sparkmagic.utils.configuration as conf
from mock import MagicMock
from nose.tools import with_setup, assert_equals, assert_raises, raises

from sparkmagic.utils.configuration import get_livy_kind
from sparkmagic.utils.constants import (
    LANGS_SUPPORTED,
    SESSION_KIND_PYSPARK,
    SESSION_KIND_SPARK,
    IDLE_SESSION_STATUS,
    BUSY_SESSION_STATUS,
    MIMETYPE_TEXT_PLAIN,
    EXPECTED_ERROR_MSG,
)
from sparkmagic.magics.sparkmagicsbase import SparkMagicBase
from sparkmagic.livyclientlib.exceptions import (
    DataFrameParseException,
    BadUserDataException,
    SparkStatementException,
)
from sparkmagic.livyclientlib.sqlquery import SQLQuery
from sparkmagic.livyclientlib.sparkstorecommand import SparkStoreCommand


def _setup():
    global magic, session, shell, ipython_display
    shell = MagicMock()
    shell.user_ns = {}
    magic = SparkMagicBase(None)
    magic.shell = shell
    session = MagicMock()
    magic.spark_controller = MagicMock()
    magic.ipython_display = MagicMock()
    conf.override_all({})


def _teardown():
    pass


def test_load_emits_event():
    spark_events = MagicMock()
    SparkMagicBase(None, spark_events=spark_events)
    spark_events.emit_library_loaded_event.assert_called_once_with()


def test_get_livy_kind_covers_all_langs():
    for lang in LANGS_SUPPORTED:
        get_livy_kind(lang)


@with_setup(_setup, _teardown)
def test_sql_df_execution_without_output_var():
    df = 0
    query = SQLQuery("")
    output_var = None
    magic.spark_controller.run_sqlquery = MagicMock(return_value=df)
    res = magic.execute_sqlquery("", None, None, None, session, output_var, False, None)

    magic.spark_controller.run_sqlquery.assert_called_once_with(query, session)
    assert res == df
    assert_equals(list(shell.user_ns.keys()), [])


@with_setup(_setup, _teardown)
def test_sql_df_execution_with_output_var():
    df = 0
    query = SQLQuery("")
    output_var = "var_name"

    magic.spark_controller = MagicMock()
    magic.spark_controller.run_sqlquery = MagicMock(return_value=df)

    res = magic.execute_sqlquery("", None, None, None, session, output_var, False, None)

    magic.spark_controller.run_sqlquery.assert_called_once_with(query, session)
    assert res == df
    assert shell.user_ns[output_var] == df


@with_setup(_setup, _teardown)
def test_sql_df_execution_quiet_without_output_var():
    df = 0
    cell = SQLQuery("")
    output_var = None

    magic.spark_controller = MagicMock()
    magic.spark_controller.run_sqlquery = MagicMock(return_value=df)

    res = magic.execute_sqlquery("", None, None, None, session, output_var, True, None)

    magic.spark_controller.run_sqlquery.assert_called_once_with(cell, session)
    assert res is None
    assert_equals(list(shell.user_ns.keys()), [])


@with_setup(_setup, _teardown)
def test_sql_df_execution_quiet_with_output_var():
    df = 0
    cell = SQLQuery("")
    output_var = "var_name"

    magic.spark_controller = MagicMock()
    magic.spark_controller.run_sqlquery = MagicMock(return_value=df)

    res = magic.execute_sqlquery("", None, None, None, session, output_var, True, None)

    magic.spark_controller.run_sqlquery.assert_called_once_with(cell, session)
    assert res is None
    assert shell.user_ns[output_var] == df


@with_setup(_setup, _teardown)
def test_sql_df_execution_quiet_with_coerce():
    df = 0
    cell = SQLQuery("", coerce=True)
    output_var = "var_name"

    magic.spark_controller = MagicMock()
    magic.spark_controller.run_sqlquery = MagicMock(return_value=df)

    res = magic.execute_sqlquery("", None, None, None, session, output_var, True, True)

    magic.spark_controller.run_sqlquery.assert_called_once_with(cell, session)
    assert res is None
    assert shell.user_ns[output_var] == df


@with_setup(_setup, _teardown)
def test_print_endpoint_info():
    current_session_id = 1
    session1 = MagicMock()
    session1.id = 1
    session1.get_row_html.return_value = """<tr><td>row1</td></tr>"""
    session2 = MagicMock()
    session2.id = 3
    session2.get_row_html.return_value = """<tr><td>row2</td></tr>"""
    magic._print_endpoint_info([session2, session1], current_session_id)
    magic.ipython_display.html.assert_called_once_with(
        """<table>
<tr><th>ID</th><th>YARN Application ID</th><th>Kind</th><th>State</th><th>Spark UI</th><th>Driver log</th><th>User</th><th>Current session?</th></tr>\
<tr><td>row1</td></tr><tr><td>row2</td></tr>\
</table>"""
    )


@with_setup(_setup, _teardown)
def test_print_empty_endpoint_info():
    current_session_id = None
    magic._print_endpoint_info([], current_session_id)
    magic.ipython_display.html.assert_called_once_with("No active sessions.")


@with_setup(_setup, _teardown)
@raises(BadUserDataException)
def test_send_to_spark_should_raise_when_variable_value_is_none():
    input_variable_name = "x_in"
    output_variable_name = "x_out"
    var_type = "str"
    max_rows = 25000
    magic.shell.user_ns[input_variable_name] = None

    magic.do_send_to_spark(
        "", input_variable_name, var_type, output_variable_name, max_rows, None
    )


@with_setup(_setup, _teardown)
@raises(BadUserDataException)
def test_send_to_spark_should_raise_when_type_is_incorrect():
    input_variable_name = "x_in"
    input_variable_value = "x_value"
    output_variable_name = "x_out"
    var_type = "incorrect"
    max_rows = 25000
    magic.shell.user_ns[input_variable_name] = input_variable_value

    magic.do_send_to_spark(
        "", input_variable_name, var_type, output_variable_name, max_rows, None
    )


@with_setup(_setup, _teardown)
def test_send_to_spark_should_print_error_when_str_command_failed():
    input_variable_name = "x_in"
    input_variable_value = "x_value"
    output_variable_name = "x_out"
    var_type = "STR"
    output_value = "error"
    max_rows = 25000
    magic.shell.user_ns[input_variable_name] = input_variable_value
    magic.spark_controller.run_command.return_value = (
        False,
        output_value,
        "text/plain",
    )

    magic.do_send_to_spark(
        "", input_variable_name, var_type, output_variable_name, max_rows, None
    )

    magic.ipython_display.send_error.assert_called_once_with(output_value)
    assert not magic.ipython_display.write.called


@with_setup(_setup, _teardown)
def test_send_to_spark_should_print_error_when_df_command_failed():
    input_variable_name = "x_in"
    input_variable_value = "x_value"
    output_variable_name = "x_out"
    var_type = "df"
    output_value = "error"
    max_rows = 25000
    magic.shell.user_ns[input_variable_name] = input_variable_value
    magic.spark_controller.run_command.return_value = (
        False,
        output_value,
        "text/plain",
    )

    magic.do_send_to_spark(
        "", input_variable_name, var_type, output_variable_name, max_rows, None
    )

    magic.ipython_display.send_error.assert_called_once_with(output_value)
    assert not magic.ipython_display.write.called


@with_setup(_setup, _teardown)
def test_send_to_spark_should_name_the_output_variable_the_same_as_input_name_when_custom_name_not_provided():
    input_variable_name = "x_in"
    input_variable_value = output_value = "x_value"
    var_type = "str"
    output_variable_name = None
    max_rows = 25000
    magic.shell.user_ns[input_variable_name] = input_variable_value
    magic.spark_controller.run_command.return_value = (True, output_value, "text/plain")
    expected_message = "Successfully passed '{}' as '{}' to Spark kernel".format(
        input_variable_name, input_variable_name
    )

    magic.do_send_to_spark(
        "", input_variable_name, var_type, output_variable_name, max_rows, None
    )

    magic.ipython_display.write.assert_called_once_with(expected_message)
    assert not magic.ipython_display.send_error.called


@with_setup(_setup, _teardown)
def test_send_to_spark_should_write_successfully_when_everything_is_correct():
    input_variable_name = "x_in"
    input_variable_value = output_value = "x_value"
    output_variable_name = "x_out"
    max_rows = 25000
    var_type = "str"
    magic.shell.user_ns[input_variable_name] = input_variable_value
    magic.spark_controller.run_command.return_value = (True, output_value, "text/plain")
    expected_message = "Successfully passed '{}' as '{}' to Spark kernel".format(
        input_variable_name, output_variable_name
    )

    magic.do_send_to_spark(
        "", input_variable_name, var_type, output_variable_name, max_rows, None
    )

    magic.ipython_display.write.assert_called_once_with(expected_message)
    assert not magic.ipython_display.send_error.called


@with_setup(_setup, _teardown)
def test_spark_execution_without_output_var():
    output_var = None

    magic.spark_controller.run_command.return_value = (True, "out", MIMETYPE_TEXT_PLAIN)
    magic.execute_spark("", output_var, None, None, None, session, None)
    magic.ipython_display.write.assert_called_once_with("out")
    assert not magic.spark_controller._spark_store_command.called

    magic.spark_controller.run_command.return_value = (
        False,
        "out",
        MIMETYPE_TEXT_PLAIN,
    )
    assert_raises(
        SparkStatementException,
        magic.execute_spark,
        "",
        output_var,
        None,
        None,
        None,
        session,
        True,
    )
    assert not magic.spark_controller._spark_store_command.called


@with_setup(_setup, _teardown)
def test_spark_execution_with_output_var():
    mockSparkCommand = MagicMock()
    magic._spark_store_command = MagicMock(return_value=mockSparkCommand)
    output_var = "var_name"
    df = "df"

    magic.spark_controller.run_command.side_effect = [
        (True, "out", MIMETYPE_TEXT_PLAIN),
        df,
    ]
    magic.execute_spark("", output_var, None, None, None, session, True)
    magic.ipython_display.write.assert_called_once_with("out")
    magic._spark_store_command.assert_called_once_with(
        output_var, None, None, None, True
    )
    assert shell.user_ns[output_var] == df

    magic.spark_controller.run_command.side_effect = None
    magic.spark_controller.run_command.return_value = (
        False,
        "out",
        MIMETYPE_TEXT_PLAIN,
    )
    assert_raises(
        SparkStatementException,
        magic.execute_spark,
        "",
        output_var,
        None,
        None,
        None,
        session,
        True,
    )


@with_setup(_setup, _teardown)
def test_spark_exception_with_output_var():
    mockSparkCommand = MagicMock()
    magic._spark_store_command = MagicMock(return_value=mockSparkCommand)
    exception = BadUserDataException("Ka-boom!")
    output_var = "var_name"
    df = "df"

    magic.spark_controller.run_command.side_effect = [
        (True, "out", MIMETYPE_TEXT_PLAIN),
        exception,
    ]
    assert_raises(
        BadUserDataException,
        magic.execute_spark,
        "",
        output_var,
        None,
        None,
        None,
        session,
        True,
    )
    magic.ipython_display.write.assert_called_once_with("out")
    magic._spark_store_command.assert_called_once_with(
        output_var, None, None, None, True
    )
    assert shell.user_ns == {}


@with_setup(_setup, _teardown)
def test_spark_statement_exception():
    mockSparkCommand = MagicMock()
    magic._spark_store_command = MagicMock(return_value=mockSparkCommand)
    exception = BadUserDataException("Ka-boom!")

    magic.spark_controller.run_command.side_effect = [
        (False, "out", "text/plain"),
        exception,
    ]
    assert_raises(
        SparkStatementException,
        magic.execute_spark,
        "",
        None,
        None,
        None,
        None,
        session,
        True,
    )
    magic.spark_controller.cleanup.assert_not_called()


@with_setup(_setup, _teardown)
def test_spark_statement_exception_shutdowns_livy_session():
    conf.override_all({"shutdown_session_on_spark_statement_errors": True})

    mockSparkCommand = MagicMock()
    magic._spark_store_command = MagicMock(return_value=mockSparkCommand)
    exception = BadUserDataException("Ka-boom!")

    magic.spark_controller.run_command.side_effect = [
        (False, "out", "text/plain"),
        exception,
    ]
    assert_raises(
        SparkStatementException,
        magic.execute_spark,
        "",
        None,
        None,
        None,
        None,
        session,
        True,
    )
    magic.spark_controller.cleanup.assert_called_once()
